import java.util.HashMap;
import java.util.Map;

/**
 * スライド辞書圧縮のサンプル
 * 
 */
public class FastSlideDictionary extends Dictionary {
	/** スライド辞書の大きさ */
	private static final int DIC_SIZE = 64 * 1024;

	@Override
	public String compress(final String input) {
		StringBuffer buffer = new StringBuffer();

		int[] candidateArray = new int[DIC_SIZE];
		Map<String, Integer> previousMap = new HashMap<String, Integer>();

		for (int cursor = 0; cursor < input.length() - 1;) {
			String key = input.substring(cursor, cursor + 2);
			int distance = 0;
			int maxLength = 0;

			Integer firstCandidate = previousMap.put(key, cursor);
			if (firstCandidate != null) {
				// 最長一致部分の探索
				for (int absCandidate = firstCandidate; cursor - absCandidate < DIC_SIZE; absCandidate = candidateArray[absCandidate
						% DIC_SIZE]) {
					// この時点で2文字一致は確定しているので比較回数を減らす
					int length = calcMatchingLength(input, absCandidate + 2, cursor + 2) + 2;

					// 一致長が最長なら保存しておく
					if (maxLength < length) {
						maxLength = length;
						distance = cursor - absCandidate;
					}
				}
			} else {
				// 無効値として使う
				firstCandidate = -DIC_SIZE;
			}
			// 次回以降の探索のために情報をアップデートする
			candidateArray[cursor % DIC_SIZE] = firstCandidate;

			String dl = String.format("[%d,%d]", distance, maxLength);
			if (dl.length() < maxLength) {
				// distanceとmaxLengthの組をファイルに出力する（圧縮）
				buffer.append(dl);
				cursor += maxLength;
			} else {
				// input.charAt(cursor)をファイルに出力する（無圧縮）
				char ch = input.charAt(cursor);
				if (ch == '[') {
					buffer.append("[[");
				} else {
					buffer.append(ch);
				}
				cursor += 1;
			}
		}
		return buffer.toString();
	}
}
